/***************************************************************************
 *   Projekt:	Entwicklung eines Lexmark 2070 Treibers fr		   *
 *              Linux als Nachverarbeitung von Ghostscript PBMs;           *
 *              zuerst mal S/W, spter vielleicht auch CMYK.               *
 *                                                                         *
 *   Autor:	Christian Kornblum, <kornblum@studst.fh-muenster.de>       *
 *                                                                         *
 *   Dank an:   Tim Norman, den Entwickler von pbm2ppa fr HP-Drucker.     *
 *              Dieses Programm orientiert sich weitgehend an seiner       *
 *              Vorlage.                                                   *
 *                                                                         *
 *   Copyright: entsprechend der GPL, unter die auch die PBM-Routinen      *
 *              fallen, die ich nur benutze.                               *
 *									   *
 *   Version:   0.6 - 29.06.1999                                           * 
 *                                                                         *
 *   Versionsgeschichte:						   *
 *    o Version 0.6                                                        *
 *         -29.06.1999 keine leeren Zeilen an Drucker senden,              *
 *                     nur noch ber tatschlich gedruckte Breite          *
 *                     drucken (schneller & professioneller)               *
 *    o Version 0.5                                                        *
 *         -seit Mai 99 im Internet verfgbar                              *
 *    o Pre-Release:                                                       *
 *         -29.03.1999 Erstes lauffhiges Release (Clean Room)             *
 *         -19.01.1999 Implementierung der pbm-Routinen                    *
 ***************************************************************************/
 
#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include "pbm.h"
#define MAX_LINES 600
#define PAGE_HEIGHT 3520
#define PAGE_WIDTH 2420
#define CARTRIDGE_PENS 104 
#define LEFT_MARGIN 10

/***************************************************************************
 *   Funktion:   lex_move                                                  *
 *   Parameter:  Zeiger auf Ausgabedatei, Pixel vertik. Versch.            *
 *   Aufgabe:    Verschiebt die Druckposition um <Pixel> 600lpi-Punkte     *
 *               nach unten.                                               *
 ***************************************************************************/
void lex_move(FILE *out, long int pixel)
{
  char buffer[] = {0x1b,0x2a,0x03,0x00,0x00};
  int i;
  buffer[4] = (char) pixel;
  buffer[3] = (char) (pixel >> 8);
  for(i=0; i<=4; fprintf(out, "%c", buffer[i++]));
}


/***************************************************************************
 *   Funktion:	lex_init						   *
 *   Parameter: Abstand vom oberen Rand in Punkten     			   *
 *   Aufgabe:   Ausgabe des Druckerheaders			           *
 ***************************************************************************/
void lex_init(FILE *out, int abstand)
{
   int i;
   char buffer[] = {0x1B,0x2A,0x80,0x1B,0x2A,0x07,
                    0x73,0x30,0x1B,0x2A,0x07,0x63};
   for(i=0; i<=11; fprintf(out, "%c", buffer[i++]));
   lex_move(out, abstand);  /* Willkrlicher Abstand vom oberen Rand */
}

/***************************************************************************
 *   Funktion:	lex_eop  						   *
 *   Parameter: Ausgabedatei, Puffer             			   *
 *   Aufgabe:   Seitenauswurf End OF Page			           *
 ***************************************************************************/
void lex_eop(FILE *out)
{
   int i;
   char buffer[] = {0x1B,0x2A,0x07,0x65,0x00,0x00,0x00,0x00};
   for(i=0; i<=7; i++) fprintf(out, "%c", buffer[i]);
}

/***************************************************************************
 *   Funktion:	print pbm						   *
 *   Parameter: Zeiger auf Eingabedatei (PBM) und Ausgabedatei 		   *
 *   Aufgabe:   Lesen des PBM und Druckabwicklung		           * 
 ***************************************************************************/
int print_pbm (FILE *in, FILE *out)
{
  char line[1024];
  pbm_stat pbm;
  int done_page, cur_height = 0, page_height = 0, numpages = 0;
  char lex_binhd[] = {0x1b,0x2a,0x04,0x00,0x00,0xFF,0xFF,
                      0x00,0x01,0x00,0x01,0x0E,0x31,0xFF,0xFF,
                      0x00,0x30,0x01,0x97,0x00,0x00,0x00,0x00,
                      0x00,0x32,0x33};
  long bytesize, stdbytesize; 
  register int i=0;
  register int cur_byte = 0;
  register int cur_pos = 0;
  char *buffer; 
  int breite, stdbreite;
  int leftmargin;
  register int redleft;
  register int redright;
  int bstart, empty_lines;

  while(make_pbm_stat(&pbm,in))
  {
    /* Begrenzung auf maximale Breite, sonst kommt nur Mist */
    if (pbm.width <= PAGE_WIDTH) stdbreite = pbm.width; 
    else stdbreite = PAGE_WIDTH;

    /* Berechnung der Puffergroesse */
    stdbytesize = 26 + 14 * stdbreite; 

    /* Speicher reservieren und leeren */
    buffer = (char *) malloc(stdbytesize - 26);
    for (i=0; i<=stdbytesize - 26; buffer[i++]=0);
    /*  Seite Laden und Drucker initialisieren... */
    lex_init(out, 230);

    /* Die Seite kann gedruckt werden... */
    done_page=0;
    page_height=0;
    cur_height=0;
    empty_lines=0;
    while(!done_page)
    {
      /* Setzen der Standardbreite vor Reduktion */
      breite = stdbreite;
      leftmargin = LEFT_MARGIN;
      bytesize = stdbytesize;
      
      /* Zeile einlesen, passenden Pen einstellen */
        pbm_readline(&pbm, line);
        cur_pos = 7 - (cur_height % 8);
        cur_byte = cur_height / 8;
        /* Hier wird binr herumgepixelt... Es klappt... */
        for(i=0; (i <= breite); i++)
	{                              
          buffer[(i * 14) + cur_byte] 
                 |= (0 < (line[i/8] & (0x80 >> (i % 8)))) << cur_pos; 
        }
        /* Genug im Buffer, also drucken */
        if ((cur_height++>=CARTRIDGE_PENS) || (page_height >= pbm.height))
	{
	  /* Reduktion der Nullbytes zur Druckoptimierung */
	  redleft = 0; redright = 0; bstart = 0;
	  while ((buffer[redleft] == 0) && (redleft < bytesize - 26))
	    redleft++;
	  while ((buffer[bytesize - 27 - redright] == 0) 
		 && (redright < bytesize - 26))
	    redright++;
	  breite -= redleft / 14 + redright / 14;
	  leftmargin += redleft / 14;
	  bstart = redleft - (redleft % 14);
	  if (bstart < 0) bstart = 0;
	  
	  /* Berechnen der Puffergre fr Drucker */
	  bytesize = 26 + 14 * breite; 

	  /* Einbauen der Puffergre in Header */  
	  lex_binhd[6] = (char) bytesize;
	  lex_binhd[5] = (char) (bytesize >> 8);
	  lex_binhd[4] = (char) (bytesize >> 16);

	  /* Einbauen der Zeilenlnge in Pixeln */
	  lex_binhd[14] = (char) breite;
	  lex_binhd[13] = (char) (breite >> 8);

	  /* Einbauen der Startkoordinaten */
	  lex_binhd[16] = (char) leftmargin;
	  lex_binhd[15] = (char) (leftmargin >> 8);
      
	  if (breite <= 0) { /* Leere Zeile, also nicht drucken */
	    empty_lines++;
	  }
	  else { /* Zeile mit Inhalt, also drucken */
	    if (empty_lines) { /* Leerzeilen als ganzes (kein Stottern) */
	      lex_move(out, empty_lines * 208);
	      empty_lines = 0;
	    }
	    for(i=0; i<26; i++) fprintf(out, "%c", lex_binhd[i]);
	    for(i=0; i<(bytesize-26);i++) fprintf(out, "%c", buffer[i+bstart]);
	    lex_move(out, 208);
	  }
          for (i=0; i<=(stdbytesize - 26); buffer[i++]=0);
          cur_height = 0; 
        }
        
        /* Seitenende durch pbm oder physikalisch */ 
        if ((page_height++>=PAGE_HEIGHT)||
            (page_height >= pbm.height)) done_page = 1; 
    }

    /* Seite auswerfen */
    lex_eop(out);

    /* eat any remaining whitespace so process will not hang */
    if(pbm.version==P1)
      fgets (line, 1024, in);

    /* Freigabe des dynamisch reservierten Speichers */
    numpages++;
    free(buffer);
  }

  if (numpages == 0)
  {
    hilfe("No pages printed!");
    return 1;
  }

  return 0;
}


/***************************************************************************
 *   Funktion:	hilfe							   *
 *   Parameter: Fehlertext						   *
 *   Aufgabe:   Ausgabe von fatalen Fehlern mit genereller Hilfe           *
 ***************************************************************************/
hilfe(char *fehler)
{
	fprintf(stderr, "Lexmark2070: %s\n\n", fehler);
}


/***************************************************************************
 *   Funktion:	main							   *
 *   Parameter: Komandozeile						   *
 *   Aufgabe:   Steuerung der Ablufe				           *
 ***************************************************************************/
int main(char argv[], int argc)
{
   return print_pbm(stdin, stdout);
}

